// Copyright © 2021 Brad Howes. All rights reserved.

import Parsing
import XCTest
import Foundation

#if canImport(Darwin)
import Darwin.C
#elseif canImport(Glibc)
import Glibc
#endif

/**
 General-purpose parser for simple math expressions made up of the common operations as well as single argument
 functions like `sqrt` and `sin` and named variables / symbols. For instance, the expression `4 * sin(t * pi)` is
 legal and references the `sin` function, the `pi` constant, and an unknown variable `t`. Parsing a legal expression
 results in an `Evaluator` that can be used to obtain Double results from the expression, such as when the
 value for `t` is known.
 */
final public class MathParser {

  /// Mapping of symbol names to an optional Double.
  public typealias SymbolMap = (String) -> Double?

  /// Mapping of names to an optional transform function
  public typealias UnaryFunctionMap = (String) -> ((Double) -> Double)?

  /// Mapping of names to an optional transform function
  public typealias BinaryFunctionMap = (String) -> ((Double, Double) -> Double)?

  /// Default symbols to use for parsing.
  public static let defaultSymbols: [String: Double] = ["pi": .pi, "π": .pi, "e": Darwin.M_E]

  /// Default 1-ary functions to use for parsing.
  public static let defaultUnaryFunctions: [String: (Double) -> Double] = [
    "sin": sin, "cos": cos, "tan": tan,
    "log10": log10, "ln": log, "loge": log, "log2": log2, "exp": exp,
    "ceil": ceil, "floor": floor, "round": round,
    "sqrt": sqrt, "cbrt": cbrt
  ]

  /// Default 2-ary functions to use for parsing.
  public static let defaultBinaryFunctions: [String: (Double, Double) -> Double] = [
    "atan2": atan2,
    "hypot": hypot,
    "pow": pow // Redundant since we support x^b expressions
  ]

  /// Symbol mapping to use during parsing and perhaps evaluation
  public let symbols: SymbolMap

  /// Function mapping to use during parsing and perhaps evaluation
  public let unaryFunctions: UnaryFunctionMap

  /// Function mapping to use during parsing and perhaps evaluation
  public let binaryFunctions: BinaryFunctionMap

  /// Parser for start of identifier (constant, variable, function)
  private lazy var identifierStart = Parse  {
    Prefix(1) { $0.isLetter }
  }

  /// Parser for remaining parts of identifier (constant, variable, function)
  private lazy var identifierRemaining = Parse {
    Prefix { $0.isNumber || $0.isLetter }
  }

  /// Parser for identifier
  private lazy var identifier = Parse {
    identifierStart
    identifierRemaining
  }.map { $0.0 + $0.1 }

  /// Type of the parser that returns a Token
  private typealias TokenParser = AnyParser<Substring, Token>

  /// Parser for a numeric constant
  private lazy var constant = Double.parser(of: Substring.self).map { Token.constant($0) }

  /// Parser for addition / subtraction operator.
  private lazy var additionOrSubtractionOperator = Parse {
    ignoreSpaces
    OneOf {
      "+".map { { tokenReducer(lhs: $0, rhs: $1, op: (+)) } }
      "-".map { { tokenReducer(lhs: $0, rhs: $1, op: (-)) } }
    }
  }

  /// Parser for valid addition / subtraction operations. This is the starting point of precedence-involved parsing.
  private lazy var additionAndSubtraction: TokenParser = LeftAssociativeInfixOperation(
    additionOrSubtractionOperator,
    higher: multiplicationAndDivision
  ).eraseToAnyParser()

  /// When true, two parsed operands in a row implies multiplication
  private let enableImpliedMultiplication: Bool

  /// Parser for multiplication / division operator.
  private lazy var multiplicationOrDivisionOperator = Parse {
    ignoreSpaces
    OneOf {
      "*".map { { tokenReducer(lhs: $0, rhs: $1, op: (*)) } }
      "/".map { { tokenReducer(lhs: $0, rhs: $1, op: (/)) } }
    }
  }

  /// Parser for valid multiplication / division operations. Higher precedence than + and -
  /// If `enableImpliedMultiplication` is `true` then one can list two operands together
  /// like `2x` and have it treated as a multiplication of `2` and the value in `x`.
  private lazy var multiplicationAndDivision = LeftAssociativeInfixOperation(
    multiplicationOrDivisionOperator,
    higher: exponentiation,
    implied: enableImpliedMultiplication ? { tokenReducer(lhs: $0, rhs: $1, op: (*)) } : nil
  )

  /// Parser for exponentiation (power) operator
  private lazy var exponentiationOperator = Parse {
    ignoreSpaces
    "^".map { { tokenReducer(lhs: $0, rhs: $1, op: (pow)) } }
  }

  /// Parser for exponentiation operation. Higher precedence than * /
  private lazy var exponentiation = LeftAssociativeInfixOperation(
    exponentiationOperator,
    higher: operand
  )

  /// Parser for a symbol. If symbol exists during parse, parser returns `.constant` token. Otherwise, parser returns
  /// `.symbol` token for later evaluation when the symbol is known.
  private lazy var symbolOrVariable = Parse {
    identifier
  }.map { (name: Substring) -> Token in
    let name = String(name)
    guard let value = self.symbols(name) else { return .variable(name) }
    return .constant(value)
  }

  /// Parser for expression in parentheses. Use Lazy due to recursive nature of this definition.
  private lazy var parenthetical = Lazy {
    "("
    self.additionAndSubtraction
    ignoreSpaces
    ")"
  }

  /// Parser for function call of 1 parameter. Use Lazy due to recursive nature of this definition.
  private lazy var functionArgs1 = Lazy {
    self.additionAndSubtraction
    ignoreSpaces
  }

  /// Parser for a single argument function call (eg `sin`).
  private lazy var function1 = Parse {
    identifier
    "("
    functionArgs1
    ")"
  }.map { (tuple) -> Token in
    let name = String(tuple.0)
    let token = tuple.1
    if let resolved = self.unaryFunctions(name) {
      // Known function name
      if case .constant(let value) = token {
        // Known constant value -- evaluate function with it and return new constant value
        return .constant(resolved(value))
      }
      return .function1(name, token)
    }

    // We don't know the function at this time, should we treat as a variable multiplying a parenthetical expression?
    return self.enableImpliedMultiplication ? tokenReducer(lhs: .variable(name), rhs: token, op: (*)) :
      .function1(name, token)
  }

  /// Parser for function call of 2 parameters. Use Lazy due to recursive nature of this definition.
  private lazy var functionArgs2 = Lazy {
    self.additionAndSubtraction
    ignoreSpaces
    ","
    ignoreSpaces
    self.additionAndSubtraction
    ignoreSpaces
  }

  /// Parser for a single argument function call (eg `atan2`).
  private lazy var function2 = Parse {
    identifier
    "("
    functionArgs2
    ")"
  }.map { (tuple) -> Token in
    let name = String(tuple.0)
    let arg1 = tuple.1.0
    let arg2 = tuple.1.1
    if let resolved = self.binaryFunctions(name),
       case .constant(let value1) = arg1,
       case .constant(let value2) = arg2 {
      // Known constant values -- evaluate function with it and return new constant value
      return .constant(resolved(value1, value2))
    }
    // Cannot reduce to a value at this time
    return .function2(name, arg1, arg2)
  }

  /// Parser for an operand of an expression. Note that order is important: a function is made up of an identifier
  /// followed by a parenthetical expression, so it must be before `parenthetical` and `symbolOrVariable`.
  private lazy var operand = Parse {
    ignoreSpaces
    OneOf {
      function2
      function1
      parenthetical
      symbolOrVariable
      constant
    }
  }

  /// Parser for a math expression. Checks that there is nothing remaining to be parsed.
  private lazy var expression = Parse {
    additionAndSubtraction
    ignoreSpaces
    End()
  }

  /**
   Construct new parser. Only supports unary functions

   - parameter symbols: optional mapping of names to constants. If not given, `defaultSymbols` will be used
   - parameter functions: optional mapping of names to 1-ary functions. If not given, `defaultUnaryFunctions` will be used
   - parameter enableImpliedMultiplication: if true treat expressions like `2π` as valid and same as `2 * π`
   */
  public init(symbols: SymbolMap? = nil,
              functions: UnaryFunctionMap? = nil,
              enableImpliedMultiplication: Bool = false) {
    self.symbols = symbols ?? { Self.defaultSymbols[$0] }
    self.unaryFunctions = functions ?? { Self.defaultUnaryFunctions[$0] }
    self.binaryFunctions = { Self.defaultBinaryFunctions[$0] }
    self.enableImpliedMultiplication = enableImpliedMultiplication
  }

  /**
   Construct new parser that supports unary and binary functions.

   - parameter symbols: optional mapping of names to constants. If not given, `defaultSymbols` will be used
   - parameter unaryFunctions: optional mapping of names to 1-ary functions. If not given, `defaultUnaryFunctions` will be used
   - parameter binaryFunctions: optional mapping of names to 2-ary functions. If not given, `defaultBinaryFunctions` will be used
   - parameter enableImpliedMultiplication: if true treat expressions like `2π` as valid and same as `2 * π`
   */
  public init(symbols: SymbolMap? = nil,
              unaryFunctions: UnaryFunctionMap? = nil,
              binaryFunctions: BinaryFunctionMap? = nil,
              enableImpliedMultiplication: Bool = false) {
    self.symbols = symbols ?? { Self.defaultSymbols[$0] }
    self.unaryFunctions = unaryFunctions ?? { Self.defaultUnaryFunctions[$0] }
    self.binaryFunctions = binaryFunctions ?? { Self.defaultBinaryFunctions[$0] }
    self.enableImpliedMultiplication = enableImpliedMultiplication
  }
}

extension MathParser {

  /**
   Parse an expression into a token that can be evaluated at a later time.

   - parameter text: the expression to parse
   - returns: optional Evaluator to use to obtain results from the parsed expression. This is nil if expression was not
   valid.
   */
  public func parse(_ text: String) -> Evaluator? {
    guard let token = try? expression.parse(text) else { return nil }
    return Evaluator(token: token, symbols: self.symbols, unaryFunctions: self.unaryFunctions,
                     binaryFunctions: self.binaryFunctions)
  }
}

/// Common expression for ignoring spaces in other parsers
fileprivate let ignoreSpaces = Skip { Optionally { Prefix { $0.isWhitespace } } }

/// All of our basic math operations reduce two inputs into one output.
fileprivate typealias Operation = (Double, Double) -> Double

/// Attempt to reduce two operand tokens and an operator. If constants, reduce to the operator applied to the constants.
/// Otherwise, return a `.mathOp` token for future evaluation.
fileprivate func tokenReducer(lhs: Token, rhs: Token, op: @escaping Operation) -> Token {
  if case let .constant(lhs) = lhs, case let .constant(rhs) = rhs {
    return .constant(op(lhs, rhs))
  }
  return .mathOp(lhs, rhs, op)
}
